/*
 *
 * UDP packet sending functions
 *
 * Large portions taken from scheduled tx tools gist
 *
 */

#define _GNU_SOURCE
#include "send_packet.h"

#include <arpa/inet.h>
#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <ifaddrs.h>
#include <inttypes.h>
#include <linux/errqueue.h>
#include <linux/ethtool.h>
#include <linux/net_tstamp.h>
#include <linux/sockios.h>
#include <net/if.h>
#include <netdb.h>
#include <netinet/in.h>
#include <poll.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "common.h"
#include "tracer.h"

static void *poll_thread(void *p);
static void process_error_queue();

static void init_tx_buffer();
static int set_if();

static int so_timestamping_flags =
    SOF_TIMESTAMPING_TX_SOFTWARE | SOF_TIMESTAMPING_SOFTWARE;

static egress_param_t *params;
static egress_stat_t *stats;
static uint64_t *kernel_latency_hist;
static int use_histogram;
static int stop_tracing;
static uint64_t packets_sent = 0;

static struct sock_txtime sk_txtime;
static char *tx_buffer;
static int sock_fd;

static int64_t timestamps_buffer[TIMESTAMP_BUFFER_SIZE];
static int ts_buf_read_index = 0;
static int ts_buf_write_index = 0;

/*
 * Init UDP socket
 */
void init_udp_send(egress_param_t *_params, egress_stat_t *_stats,
                   int _use_histogram, int _stop_tracing, uint64_t *_kernel_latency_hist) {
  int set_if_err;
  pthread_t thread;

  params = _params;
  stats = _stats;
  kernel_latency_hist = _kernel_latency_hist;
  use_histogram = _use_histogram;
  stop_tracing = _stop_tracing;

  init_tx_buffer();

  sock_fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (sock_fd < 0) error(EXIT_FAILURE, errno, "Socket creation failed\n");

  set_if_err = set_if();
  if (set_if_err < 0) error(EXIT_FAILURE, errno, "Couldn't set interface\n");

  if (setsockopt(sock_fd, SOL_SOCKET, SO_PRIORITY, &params->packet_priority,
                 sizeof(params->packet_priority)))
    error(EXIT_FAILURE, errno, "Couldn't set socket priority\n");

  if (setsockopt(sock_fd, SOL_SOCKET, SO_BINDTODEVICE, params->network_if,
                 strlen(params->network_if)))
    error(EXIT_FAILURE, errno, "setsockopt SO_BINDTODEVICE failed\n");

  if (params->use_etf) {
    sk_txtime.clockid = CLOCK_REALTIME;
    sk_txtime.flags = SOF_TXTIME_REPORT_ERRORS;

    if (setsockopt(sock_fd, SOL_SOCKET, SO_TXTIME, &sk_txtime,
                   sizeof(sk_txtime)))
      error(EXIT_FAILURE, errno, "setsockopt SO_TXTIME failed\n");
  }

  if (params->use_timestamps) {
    if (setsockopt(sock_fd, SOL_SOCKET, SO_TIMESTAMPING, &so_timestamping_flags,
                   sizeof(so_timestamping_flags)))
      error(EXIT_FAILURE, errno, "setsockopt SO_TIMESTAMPING failed\n");
  }

  // Create poll thread
  if (pthread_create(&thread, NULL, poll_thread, NULL))
    error(EXIT_FAILURE, errno, "Couldn't create poll thread");
}


/*
 * Sends udp packets
 */
void send_udp_packet(char *data, uint64_t txtime) {
  struct msghdr msg;     // Message hardware, sent to the socket
  struct cmsghdr *cmsg;  // Control message hardware, for txtime
  char control[CMSG_SPACE(sizeof(txtime))] = {};  // Stores txtime
  struct iovec iov;  // The iovec structures stores the TX buffer
  int sendmsgerr;
  struct sockaddr_in sin;  // Server address
  struct timespec ts;      // timestamp for userspace timestamping

  if (params->use_timestamps) {
    clock_gettime(CLOCK_REALTIME, &ts);
    timestamps_buffer[ts_buf_write_index] = ts_to_uint(ts);
    ts_buf_write_index = (ts_buf_write_index + 1) % TIMESTAMP_BUFFER_SIZE;
  }

  packets_sent++;

  for(int i = 0; i < (int)params->tx_buffer_len; i++)
    tx_buffer[i] = data[i];

  memset(&sin, 0, sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = inet_addr(params->server_ip);
  sin.sin_port = htons(SERVER_PORT_INT);

  iov.iov_base = tx_buffer;
  iov.iov_len = params->tx_buffer_len;

  memset(&msg, 0, sizeof(msg));
  msg.msg_name = &sin;
  msg.msg_namelen = sizeof(sin);
  msg.msg_iov = &iov;
  msg.msg_iovlen = 1;

  if (params->use_etf) {
    // We specify the transmission time in the CMSG.
    msg.msg_control = control;
    msg.msg_controllen = sizeof(control);

    cmsg = CMSG_FIRSTHDR(&msg);
    cmsg->cmsg_level = SOL_SOCKET;
    cmsg->cmsg_type = SCM_TXTIME;
    cmsg->cmsg_len = CMSG_LEN(sizeof(uint64_t));
    *((uint64_t *)CMSG_DATA(cmsg)) = txtime;
    msg.msg_controllen = cmsg->cmsg_len;
  }

  sendmsgerr = sendmsg(sock_fd, &msg, 0);
  if (sendmsgerr < 0)
    error(EXIT_FAILURE, errno, "sendmsg failed, ret value: %d\n", sendmsgerr);
}

static void *poll_thread(void *p) {
  (void)p;
  // Poll file descriptor
  struct pollfd poll_fd = {.fd = sock_fd};

  while (1) {
    int ret;

    ret = poll(&poll_fd, 1, -1);
    if (ret == 1 && poll_fd.revents & POLLERR) {
      process_error_queue();
    }
  }

  return NULL;
}

static void process_error_queue() {
  int recv_ret;

  // IO vector
  unsigned char data_buffer[256];  // Buffer in io vector
  struct iovec iov = {.iov_base = data_buffer, .iov_len = sizeof(data_buffer)};

  // Control data, will store error or timestamps
  unsigned char msg_control[CMSG_SPACE(sizeof(struct sock_extended_err)) +
                            CMSG_SPACE(sizeof(struct timespec))];

  // Message hardware structure, containts IO vector and control message
  // hardware
  struct msghdr msg = {.msg_iov = &iov,
                       .msg_iovlen = 1,
                       .msg_control = msg_control,
                       .msg_controllen = sizeof(msg_control)};

  struct cmsghdr *cmsg;

  // Timestamps and errors are received in the error queue
  recv_ret = recvmsg(sock_fd, &msg, MSG_ERRQUEUE | MSG_DONTWAIT);
  if (recv_ret == -1) {
    fprintf(stderr, "recvmsg() failed\n");
    return;
  }

  for (cmsg = CMSG_FIRSTHDR(&msg); cmsg; cmsg = CMSG_NXTHDR(&msg, cmsg)) {
    // If a timestamp was received
    if (cmsg->cmsg_level == SOL_SOCKET && cmsg->cmsg_type == SO_TIMESTAMPING) {
      struct timespec *stamp = (struct timespec *)CMSG_DATA(cmsg);

      int kernel_latency =
          (ts_to_uint(*stamp) - timestamps_buffer[ts_buf_read_index]) / 1000;
      ts_buf_read_index = (ts_buf_read_index + 1) % TIMESTAMP_BUFFER_SIZE;

      stats->min_kernel_latency =
          _min_(kernel_latency, stats->min_kernel_latency);
      stats->max_kernel_latency =
          _max_(kernel_latency, stats->max_kernel_latency);
      stats->avg_kernel_latency =
          (stats->max_kernel_latency * packets_sent + kernel_latency) /
          (packets_sent + 1);

      if (use_histogram) {
        if (kernel_latency > MAX_KERNEL_LATENCY)
          stats->high_kernel_latency++;
        else
          kernel_latency_hist[kernel_latency]++;
      }

    }
    // If an error was received
    else {
      struct sock_extended_err *serr = (void *)CMSG_DATA(cmsg);

      if (serr->ee_origin != SO_EE_ORIGIN_TXTIME) continue;

      switch (serr->ee_code) {
        case SO_EE_CODE_TXTIME_INVALID_PARAM:
          stats->invalid_parameter++;
          break;
        case SO_EE_CODE_TXTIME_MISSED:
          if(stop_tracing) {
            tracemark("Deadline missed\n");
            tracing(0);
            exit(EXIT_FAILURE);
          }
          stats->missed_deadline++;
          break;
        default:
          fprintf(stderr, "Uknown TxTime error\n");
      }
    }
  }
}

// Sets the interface
static int set_if() {
  struct ifreq ifreq;

  memset(&ifreq, 0, sizeof(ifreq));
  strncpy(ifreq.ifr_name, params->network_if, sizeof(ifreq.ifr_name) - 1);

  if (ioctl(sock_fd, SIOCGIFINDEX, &ifreq))
    error(EXIT_FAILURE, errno, "ioctl SIOCGIFINDEX failed\n");

  return ifreq.ifr_ifindex;
}

static void init_tx_buffer() {
  if (params->tx_buffer_len < 1) {
    fprintf(stderr, "tx buffer length should be greater than 1\n");
    exit(EXIT_FAILURE);
  }

  tx_buffer = malloc(params->tx_buffer_len);
}

