#!/usr/bin/env python3
#
# Copyright (c) 2018, Intel Corporation
#
# SPDX-License-Identifier: BSD-3-Clause
#

# Expected input file format is a CSV file with:
#
# <FRAME_NUMBER, FRAME_ARRIVAL_TIME, FRAME_PAYLOAD_BYTES>
# E.g.:
# 1,1521534608.000000456,00:38:89:bd:a1:93:1d:15:(...)
# 2,1521534608.001000480,00:38:89:bd:a1:93:1d:15:(...)
#
# Frame number:         sequence number for each frame
# Frame arrival time:   Rx HW timestamp for each frame
# Frame Payload:        payload starting with 64bit timestamp (txtime)
#
# This can be easily generated with tshark with the following command line:
#  $ tshark -r CAPTURE.pcap -t e -E separator=, -T fields -e frame.number \
#           -e frame.time_epoch \
#           -e data.data > DATA.out
#
import argparse
import csv
import struct
import math
import sys
import json

def compute_offsets_stats(file_path, interval):

    interval = int(interval)

    measure_set = {}
    measure_set["measure_type"] = "packet_jitter_tcpdump"
    measure_set["props_names"] = ["jitter"]
    measure_set["units"] = ["us"]
    measure_set["props_type"] = "histogram"
    measure_set["metadata"] = {"i": interval}
    measure_set["middle"] = 1000

    histogram = [0 for i in range(0, 2000)]

    with open(file_path) as f:
        min_t = sys.maxsize
        max_t = -sys.maxsize

        i = 0
        prev_tstamp = 0

        for line in csv.reader(f):
            arrival_tstamp = int(line[1].replace('.', ''))

            if i > 0:
                val = arrival_tstamp - prev_tstamp
                jitter = (val - interval * 1000) / 1000
                histogram_index = int(1000 + jitter)
                if histogram_index < 0 or histogram_index > 2000:
                    print("jitter too high: {}\n".format(jitter))
                else:
                    histogram[histogram_index] += 1

                # Update statistics.
                min_t = jitter if jitter < min_t else min_t
                max_t = jitter if jitter > max_t else max_t

            i += 1
            prev_tstamp = arrival_tstamp

        if i > 0:
            print("Jitter: {}us".format(max_t - min_t))

    measure_set["props"] = [histogram]

    data = {}
    data["measure_sets"] = [measure_set]

    path = "tcpdump_histogram"

    with open(path, 'w') as outfile:
        json.dump(data, outfile)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', dest='file_path', default=None, type=str,
                       help='Path to input file (e.g. DATA.out) generated by tshark with:\
                       tshark -r CAPTURE.pcap -t e -E separator=, -T\
                       fields -e frame.number -e frame.time_epoch\
                       -e data.data > DATA.out')
    parser.add_argument('-i', nargs=1, required=False, help='Indicate which interval was used')

    # Default interval
    interval = 1000000

    args = parser.parse_args()

    if args.i is not None:
        interval = args.i[0]

    if args.file_path is not None:
        compute_offsets_stats(args.file_path, interval)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()

